// -*- mode: js2; indent-tabs-mode: nil; js2-basic-offset: 4 -*-

const Gio = imports.gi.Gio;
const GLib = imports.gi.GLib;
const Meta = imports.gi.Meta;
const Clutter = imports.gi.Clutter;
const Gi = imports._gi;
const Panel = imports.ui.panel.Panel;
const Layout = imports.ui.layout;
const OverviewControls = imports.ui.overviewControls;
const Workspace = imports.ui.workspace;
const Main = imports.ui.main;

const Gettext = imports.gettext.domain('gnome-shell-extensions');
const _ = Gettext.gettext;

const ExtensionUtils = imports.misc.extensionUtils;

const CLASSIC_OVERRIDES_SCHEMA = 'org.gnome.shell.extensions.classic-overrides';

const WINDOW_CLONE_MAXIMUM_SCALE = 1.0;

function isSLEClassicMode() {
    return Main.sessionMode.currentMode == "sle-classic" ? true : false;
}

// panel.js: Replace the original "Panel._allocate".
//
// The differences are quite subtle but yet pervasive, so despite the
// similarity, there will be two versions of allocation functions.
//
// The main difference is to use the empty(in Classic) _centerBox for window
// list.
//
// NOTE: has to be a *direct* replacement to have proper `this` setup
function sle_classic_allocate (box, flags) {
    // NOTE: for convenience, the followings are shared. The params of this
    // function is also shared.
    let allocWidth = box.x2 - box.x1;
    let allocHeight = box.y2 - box.y1;
    let [leftMinWidth, leftNaturalWidth] = this._leftBox.get_preferred_width(-1);
    let [centerMinWidth, centerNaturalWidth] = this._centerBox.get_preferred_width(-1);
    let [rightMinWidth, rightNaturalWidth] = this._rightBox.get_preferred_width(-1);

    // bind sub function's `this` to the current `this`, i.e. Main.panel
    let allocateSC = _allocateSC.bind(this);
    // For convenience, defined here to allow access to local vars
    function _allocateSC () {
        let isTextDirRTL = ( this.get_text_direction() === Clutter.TextDirection.RTL );
        // A mask to create allocation.
        let maskBox = new Clutter.ActorBox();

        // NOTE: policy here is to satisfy left&right boxes space needs first with
        // center box can be shrinked. (This is the opposite of the original
        // allocator, this idea here is to accomodate SLE-Classic window list within
        // center box first. )
        let maxSideWidth = Math.floor( (allocWidth - centerMinWidth) / 2 );
        let leftWidth = Math.min(maxSideWidth, leftNaturalWidth);
        let rightWidth = Math.min(maxSideWidth, rightNaturalWidth);
        // NOTE: in the case, flooring for maxSideWidth does happen, the following equation yields:
        // : centerWidth = centerMinWidth + 1
        // this avoids the cumbersome floor/ceil in orginal code.
        //
        // `centerWidth` is shrinkable with an lower bound `centerMinWidth` or `centerMinWidth + 1`
        let centerWidth = allocWidth - leftWidth - rightWidth;

        // Adjust left/center/right boxes start Y position
        //
        // After moving the panel to the bottom side, the top border will be
        // added instead of the bottom border. However all 3 box of the main
        // panel are absolutely positioned starting from the top edge of the
        // panel originally, this results that the bottom pixel line is not
        // clickable. Add a Y offset to fix it.
        //
        // NOTE: we can set y1 as (MainLayoutManager.primaryMonitor.height -
        // allocHeight), which seems to be easier, however the solution below is
        // more generic and versatile.
        let node = Main.panel.get_theme_node();
        let borderTopWidth = node.get_length("border-top");
        let y_offset = borderTopWidth;

        maskBox.y1 = y_offset;
        // TODO: in previous version of SLE Classic the following is
        // `allocHeight + y_offset`, why?
        maskBox.y2 = allocHeight;

        // Normal order, from left to right
        if ( ! isTextDirRTL ) {
            maskBox.x1 = 0;
            maskBox.x2 = maskBox.x1 + leftWidth;
            this._leftBox.allocate(maskBox);

            maskBox.x1 = maskBox.x2;
            maskBox.x2 = maskBox.x1 + centerWidth;
            this._centerBox.allocate(maskBox);

            maskBox.x1 = maskBox.x2;
            maskBox.x2 = maskBox.x1 + rightWidth;
            this._rightBox.allocate(maskBox);
        }
        // Right to Left: essentially right box left box are swapped
        else {
            maskBox.x2 = allocWidth;
            maskBox.x1 = maskBox.x2 - leftWidth;
            this._leftBox.allocate(maskBox);

            maskBox.x2 = maskBox.x1;
            maskBox.x1 = maskBox.x2 - centerWidth;
            this._centerBox.allocate(maskBox);

            maskBox.x2 = maskBox.x1;
            maskBox.x1 = maskBox.x2 - rightWidth;
            this._rightBox.allocate(maskBox);
        }
    }

    allocateSC();

    // TODO: what are these corners for? They are zero on my test machines.
    // Without proper understanding of their purposes, they are the same for
    // both original code and SLE Classic.
    let cornerMinWidth, cornerMinHeight;
    let cornerWidth, cornerHeight;
    let childBox = new Clutter.ActorBox();

    [cornerMinWidth, cornerWidth] = this._leftCorner.get_preferred_width(-1);
    [cornerMinHeight, cornerHeight] = this._leftCorner.get_preferred_height(-1);
    childBox.x1 = 0;
    childBox.x2 = cornerWidth;
    childBox.y1 = allocHeight;
    childBox.y2 = allocHeight + cornerHeight;
    this._leftCorner.allocate(childBox);

    [cornerMinWidth, cornerWidth] = this._rightCorner.get_preferred_width(-1);
    [cornerMinHeight, cornerHeight] = this._rightCorner.get_preferred_height(-1);
    childBox.x1 = allocWidth - cornerWidth;
    childBox.x2 = allocWidth;
    childBox.y1 = allocHeight;
    childBox.y2 = allocHeight + cornerHeight;
    this._rightCorner.allocate(childBox);
}

// layout.js: Replace the origin "box.set_position" call
function setMainPanelPosition (mainPanel, primaryMonitor) {
    let mainPanelHeight = mainPanel.height === 0 ? 28 : mainPanel.height;

    mainPanel.set_position(primaryMonitor.x,
                           primaryMonitor.y + primaryMonitor.height - mainPanelHeight);

}

function _updateBoxesNew() {
    this.screenShieldGroup.set_position(0, 0);
    this.screenShieldGroup.set_size(global.screen_width, global.screen_height);

    if (!this.primaryMonitor)
        return;

    setMainPanelPosition(this.panelBox, this.primaryMonitor);
    this.panelBox.set_size(this.primaryMonitor.width, -1);

    this.keyboardIndex = this.primaryIndex;
}

//bsc#1064378
function _vfunc_allocateNew(container, box) {
    const childBox = new Clutter.ActorBox();
    const [width, height] = box.get_size();
    childBox.set_size(width, height - Main.layoutManager.panelBox.height);
    Main.overview._overview._controls.layout_manager.vfunc_allocate(container, childBox);
}

//bsc#1064378
function computeScaleAndSpaceNew(layout) {
    let area = layout.area;

    let hspacing = (layout.maxColumns - 1) * this._columnSpacing;
    //change layout.numRows to start from 1
    let vspacing = layout.numRows * this._rowSpacing;

    let spacedWidth = area.width - hspacing;
    let spacedHeight = area.height - vspacing;

    let horizontalScale = spacedWidth / layout.gridWidth;
    let verticalScale = spacedHeight / layout.gridHeight;

    // Thumbnails should be less than 70% of the original size
    let scale = Math.min(horizontalScale, verticalScale, WINDOW_CLONE_MAXIMUM_SCALE);

    let scaledLayoutWidth = layout.gridWidth * scale + hspacing;
    let scaledLayoutHeight = layout.gridHeight * scale + vspacing;
    let space = (scaledLayoutWidth * scaledLayoutHeight) / (area.width * area.height);

    layout.scale = scale;
    layout.space = space;
}

//bsc#1064378
function computeWindowSlotsNew(layout, area) {
    this._computeRowSizes(layout);

    let { rows: rows, scale: scale } = layout;

    let slots = [];

    // Do this in three parts.
    let heightWithoutSpacing = 0;
    for (let i = 0; i < rows.length; i++) {
        let row = rows[i];
        heightWithoutSpacing += row.height;
    }

    //change rwos.length to start from 1
    let verticalSpacing = rows.length * this._rowSpacing;
    let additionalVerticalScale = Math.min(1, (area.height - verticalSpacing) / heightWithoutSpacing);

    // keep track how much smaller the grid becomes due to scaling
    // so it can be centered again
    let compensation = 0;
    let y = 0;

    for (let i = 0; i < rows.length; i++) {
        let row = rows[i];

        // If this window layout row doesn't fit in the actual
        // geometry, then apply an additional scale to it.
        let horizontalSpacing = (row.windows.length - 1) * this._columnSpacing;
        let widthWithoutSpacing = row.width - horizontalSpacing;
        let additionalHorizontalScale = Math.min(1, (area.width - horizontalSpacing) / widthWithoutSpacing);

        if (additionalHorizontalScale < additionalVerticalScale) {
            row.additionalScale = additionalHorizontalScale;
            // Only consider the scaling in addition to the vertical scaling for centering.
            compensation += (additionalVerticalScale - additionalHorizontalScale) * row.height;
        } else {
            row.additionalScale = additionalVerticalScale;
            // No compensation when scaling vertically since centering based on a too large
            // height would undo what vertical scaling is trying to achieve.
        }

        row.x = area.x + (Math.max(area.width - (widthWithoutSpacing * row.additionalScale + horizontalSpacing), 0) / 2);
        row.y = area.y + (Math.max(area.height - (heightWithoutSpacing + verticalSpacing), 0) / 2) + y;
        y += row.height * row.additionalScale + this._rowSpacing;
    }

    compensation = compensation / 2;

    for (let i = 0; i < rows.length; i++) {
        let row = rows[i];
        let x = row.x;
        for (let j = 0; j < row.windows.length; j++) {
            let window = row.windows[j];

            let s = scale * this._computeWindowScale(window) * row.additionalScale;
            let cellWidth = window.width * s;
            let cellHeight = window.height * s;

            s = Math.min(s, WINDOW_CLONE_MAXIMUM_SCALE);
            let cloneWidth = window.width * s;

            let cloneX = x + (cellWidth - cloneWidth) / 2;
            let cloneY = row.y + row.height * row.additionalScale - cellHeight + compensation;

            // Align with the pixel grid to prevent blurry windows at scale = 1
            cloneX = Math.floor(cloneX);
            cloneY = Math.floor(cloneY);

            slots.push([cloneX, cloneY, s, window]);
            x += cellWidth + this._columnSpacing;
        }
    }
    return slots;
}

let _updateBoxesOrigin;
let computeScaleAndSpaceOrigin;
let computeWindowSlotsOrigin;
let layoutVfuncAllocate;
let heightChangeId;
let _origVfuncAllocate;

function init() {
    ExtensionUtils.initTranslations();

    _updateBoxesOrigin = Main.layoutManager._updateBoxes;

    //bsc#1064378
    computeScaleAndSpaceOrigin = Workspace.LayoutStrategy.prototype.computeScaleAndSpace;
    computeWindowSlotsOrigin = Workspace.LayoutStrategy.prototype.computeWindowSlots;
}

function enable() {
    if (!isSLEClassicMode())
        return;

    _origVfuncAllocate = Main.panel.__proto__.vfunc_allocate;

    Main.panel.__proto__[Gi.hook_up_vfunc_symbol]('allocate',  (box, flags) => {
        Main.panel.vfunc_allocate.call(Main.panel, box, flags);
        sle_classic_allocate.call(Main.panel, box, flags);
    });

    Main.layoutManager._updateBoxes = _updateBoxesNew;

    //bsc#1064378
    layoutVfuncAllocate = Main.overview._overview._controls.layout_manager.__proto__.vfunc_allocate;
    Main.overview._overview._controls.layout_manager.__proto__[Gi.hook_up_vfunc_symbol]('allocate', _vfunc_allocateNew);

    //bsc#1064378
    Workspace.LayoutStrategy.prototype.computeScaleAndSpace = computeScaleAndSpaceNew;
    Workspace.LayoutStrategy.prototype.computeWindowSlots = computeWindowSlotsNew;

    //bsc#1046570
    heightChangeId = Main.panel.connect('notify::height', () => {
        let primaryMonitor = Main.layoutManager.primaryMonitor;
        Main.layoutManager.panelBox.set_position(primaryMonitor.x,
            primaryMonitor.y + primaryMonitor.height - Main.panel.height);
    });

    Main.layoutManager._monitorsChanged();
}

function disable() {
    let primaryMonitor = Main.layoutManager.primaryMonitor;
    Main.layoutManager.panelBox.set_position(primaryMonitor.x, primaryMonitor.y);

    if (_origVfuncAllocate) {
        Main.panel.__proto__[Gi.hook_up_vfunc_symbol]('allocate', _origVfuncAllocate);
        _origVfuncAllocate = null;
    }

    Main.layoutManager._updateBoxes = _updateBoxesOrigin;

    //bsc#1064378
    Main.overview._overview._controls.layout_manager.__proto__[Gi.hook_up_vfunc_symbol]('allocate', layoutVfuncAllocate);

    //bsc#1064378
    Workspace.LayoutStrategy.prototype.computeScaleAndSpace = computeScaleAndSpaceOrigin;
    Workspace.LayoutStrategy.prototype.computeWindowSlots = computeWindowSlotsOrigin;

    //bsc#1046570
    if (heightChangeId)
        Main.panel.disconnect(heightChangeId);
}
